// ============================================================================
// E:\aUE4\FBO_CBP_TurbineSim
// CBP 13-11-21
// Based on Dean Sisman's code
// Rev 14-11-21 Restore of UDK log system. Removal of Tick and use of Timer
// fixed beta reset bug for region 2-3 transition
// 17-11-21 Region3 Kp Kd exposed. Entering Region-3 Message Added
// ============================================================================

#include "Turbine.h"
#include "Blade.h"

//==========================================================================================================================================
//	INITIALISING
//==========================================================================================================================================

//initializes required vals for the engine
ATurbine::ATurbine() {
	//PrimaryActorTick.bCanEverTick = true;
	PrimaryActorTick.bCanEverTick = false;
}


// Called when the game starts or when spawned
void ATurbine::BeginPlay() {
	Super::BeginPlay();

	//set vars to inputted initial vals
	beta = region2Beta;
	omega = omegaInit;
	windV = windVInit;

	omegaRated = tsrOpt*vRated/R;
	deltaBeta = 0.0;


	SpawnBlades();
	ReadCpLambdaFile();	

	UE_LOG(LogTemp, Warning, TEXT("Output: %e"), omegaRated);

	GetWorldTimerManager().SetTimer(TimerHandle, this,
		&ATurbine::ComputeDT, dT, true);
	itn = 0;

	GEngine->AddOnScreenDebugMessage(-1, 5.0f, FColor::Red,
		FString::Printf(TEXT("Engine Release 26-11-21.1"), dT));

	//GEngine->AddOnScreenDebugMessage(-1, 40.0f, FColor::Yellow,
 	//	FString::Printf(TEXT("Starting: dT = %e"), dT));

	region3MessageShown = false;
}


//spawn in the turbines blades
void ATurbine::SpawnBlades() {
	auto world = GetWorld();
	FActorSpawnParameters params;
	params.Owner = this;
	for (int i = 0; i < 3; i++) {
		FVector bladePos = GetActorLocation() + FVector(0,50,830);
		FRotator bladeRot = FRotator((i * 120), 0, 0);
		blades.Add(world->SpawnActor<ABlade>(ABlade::StaticClass(), bladePos, bladeRot, params));
	}
}


//populate the cpl array with data from external text document
void ATurbine::ReadCpLambdaFile() {
	//import data from txt file
	FString importedString;
	FString file = FPaths::ProjectConfigDir();
	file.Append(TEXT("CpLambda.txt"));
	FFileHelper::LoadFileToString(importedString, *file);

	//split by lines and commas and add data to 2d array
	TArray<FString> parsedString;
	importedString.ParseIntoArray(parsedString, TEXT("\n"), true); 
	for (int i = 0; i < parsedString.Num(); i++) {
		CplArray.Add(FCplArray());
		TArray<FString> parsedLine;
		FString line = parsedString[i];
		line.ParseIntoArray(parsedLine, TEXT(","), true); 
		for (int j = 0; j < parsedLine.Num(); j++) CplArray[i].Add(FCString::Atod(*parsedLine[j]));
		// GEngine->AddOnScreenDebugMessage(-1, 40.0f, FColor::Yellow,
		// 		FString::Printf(TEXT("Added val: %f"), CplArray[i][j]));
	}
}


//==========================================================================================================================================
//	TICK METHOD Replaced by TIMER CBP 14-11-21
//==========================================================================================================================================

void ATurbine::ComputeDT() {
	if (bRunning) {
		for (int i = 0; i < speedUp; i++) Computation();
		Visualization(dT * scaling);
		LogData();
		itn++;
	}
	if (time >= stopTime) {
		if (DEBUG) GEngine->AddOnScreenDebugMessage(-1, 4.0f, FColor::Red,
			FString::Printf(TEXT(">>>> SIMULATION STOPPED <<<<<")));
		bRunning = false;
	}
}


void ATurbine::LogData() {
	int m;
	m = (int(logInterval / dT));
	//GEngine->AddOnScreenDebugMessage(-1, 40.0f, FColor::Blue,
		//FString::Printf(TEXT(">>>> LogData: time = %e,   dT = %e, logInterval = %e, m = %d"), time, dT, logInterval,m));
	if (itn % m == 0) {
		//if(DEBUG) GEngine->AddOnScreenDebugMessage(1, 4.0f, FColor::Blue,
		//	FString::Printf(TEXT(">>>> SNAPSHOT: time=  %e,  omega = %e, Cp:  %e,    windV:  %e"), time, omega,Cp, windV));
		snapshots.Add(FSnapshot{ time,windV,omega,Cp,power,tsr,beta });
	}
		
}


//==========================================================================================================================================
//	COMPUTATION
//==========================================================================================================================================

void ATurbine::Computation() {

	// assume turbine is not starting from rest: omega = omegaInit in BeginPlay()
	//CODERS
	// restart turbine if stops
	if (omega < 0.000001) {
		omega = omegaInit;
		GEngine->AddOnScreenDebugMessage(1, 1.0f, FColor::Red,
				FString::Printf(TEXT(">>>>>>> Turbine Restart")));
	}

	// clamp windV to minimum;
	if (windV < 0.01) windV = 0.01;

	//step the wind if option selected
	//if (useSteppedWindSpeed) windV = SteppedWindSpeed();
	// MODS FOR CODERS
	if(windType == stepped) windV = SteppedWindSpeed();
	if(windType == series && itn % 100 == 0)  windV = 5.1 + CalcWindSpeed(averageWindSpeed);
	
	torqueB = 0.0;

	// set region ----------------------------------------------------------------
	prevRegion = region;

	if (windV < vCutIn) region = 1;
	else if (windV > vCutOut) region = 4;
	else if (windV > vCutIn && omega < omegaRated) {
		region = 2;
	} 
	else {
		if (!region3MessageShown) {
			//GEngine->AddOnScreenDebugMessage(-1, 5.0f, FColor::Red,
			//	FString::Printf(TEXT(">>>>>>>>>>> Entering Region-3 <<<<<<<<<<<<<<<<<<")));
			region3MessageShown = true;
		}
		region = 3;
	}
	//MODS FOR CODERS
	//if(omega > omegaRated)
	//	GEngine->AddOnScreenDebugMessage(-1, 2.0f, FColor::Red,
	//		FString::Printf(TEXT(">>>>>>>>>>> omega > omegaRated")));

	//R=20;
	//MODS FOR CODERS
	if(region != prevRegion)
		GEngine->AddOnScreenDebugMessage(1, 2.0f, FColor::Yellow,
			FString::Printf(TEXT(">>>> Transit from Region %d to region %d <<<<<<<<<<<"),prevRegion,region));

	// FSM run according to active region-----------------------------------------
	switch(region){
		case 1:
			//add
			break;
		case 2:
			Region2();
			// MODS FOR CODERS
			// Make sure beta returns to Region2 value
			beta += 10*(region2Beta - beta) * dT;
			break;
		case 3:
			deltaBeta = Kp*(omega - omegaRated) *dT;
    		beta += deltaBeta;
			Region2();
			break;
		case 4:
			// This is a brutal dead stop, need to have braking effect here so omega
			// and power go gracefully to zero
			// CODERS
			omega -= omega * dT / 5.0;  //Time constant 5 sec
			//omega = 0.0;
			//power = 0.0;
			//Region2();
			break;
	}

	time += dT;
}


double ATurbine::SteppedWindSpeed() {
	double newWindV;
	//GEngine->AddOnScreenDebugMessage(-1, 40.0f, FColor::Blue,
	//	FString::Printf(TEXT("SteppedWindSpeed() stepTime = %e"), stepTime));
	if (stepTime >= windStepTimeInterval) {
		stepTime = 0;
		newWindV = windV + windStepHeight;
		//GEngine->AddOnScreenDebugMessage(-1, 40.0f, FColor::Yellow,
		//	FString::Printf(TEXT("time:%e,  windV=%e"),time,W));
		GEngine->AddOnScreenDebugMessage(-1, 4.0f, FColor::Blue,
			FString::Printf(TEXT("SteppedWindSpeed() time:  %e,   Cp:  %e, beta = %e   windV:  %e"), time, Cp, beta, newWindV));
		return newWindV;
	}

	stepTime += dT;
	return windV;
}


//standard region 2 operation
void ATurbine::Region2() {
 	tsr = omega*R/windV;
	Cp = LookUpInCpLambdaArray(tsr,beta);
	//GEngine->AddOnScreenDebugMessage(-1, 40.0f, FColor::Yellow,
	//	FString::Printf(TEXT("time:%e   Cp:%e"), time,Cp));

	torqueW = (0.5*rho*pi*(PowerOf(R, 5)) * (Cp    / PowerOf(tsr,    3)) * PowerOf(omega,2));
	if (tsr == 0) torqueW = 0;

	torqueG = (0.5*rho*pi*(PowerOf(R, 5)) * (cpMax / PowerOf(tsrOpt, 3)) * PowerOf(omega,2));

	error =  (torqueW - torqueG) - torqueB;
	omega += error*dT/J;
	if (omega < 0) omega = 0;
	theta += omega*dT;
	power = torqueG*omega;

	//normalise
	if(theta > pi) theta= theta - 2*pi;
	if(theta < -pi) theta = theta + 2*pi;
	//if (theta > 2*pi) theta = theta - 2 * pi;
	//if (theta < -2*pi) theta = theta + 2 * pi;
}


double ATurbine::Interpolate(double beta_, double beta1_, double beta2_, double Cp1_, double Cp2_)  {
  double ret;
  ret = (Cp2_ - Cp1_)/(beta2_ - beta1_)*(beta_ - beta1_) + Cp1_;
  return ret;
}


double ATurbine::LookUpInCpLambdaArray(double tsr_, double pitch_)  { 
  int row,col;
  double b1,b2,Cp1,Cp2;
  double Cpp;

  row = (int)((tsr_ - 3.0) * 10);
  if (row > 90) row = 90;
  if (row < 0) row = 0;

  // catch the limits of the pitch_

  if( pitch_ < 3.2) col = 0;
  if( pitch_ > 33.7) col = 12;

  // now do the interpolations.

  b1 = 3.2;
  b2 = 3.7;
  if( (pitch_ > b1)&&(pitch_ <= b2) ) {
	col = 0;
	Cp1 = CplArray[row][col];
	col = 1;
	Cp2 = CplArray[row][col];
    Cpp = Interpolate(pitch_,b1,b2,Cp1,Cp2);
	return Cpp;
  }

  b1 = 3.7;
  b2 = 4.2;
  if( (pitch_ > b1)&&(pitch_ <= b2) ) {
	col = 1;
	Cp1 = CplArray[row][col];
	col = 2;
	Cp2 = CplArray[row][col];
    Cpp = Interpolate(pitch_,b1,b2,Cp1,Cp2);
	return Cpp;
  }

  b1 = 4.2;
  b2 = 4.7;
  if( (pitch_ > b1)&&(pitch_ <= b2) ) {
	col = 2;
	Cp1 = CplArray[row][col];
	col = 3;
	Cp2 = CplArray[row][col];
    Cpp = Interpolate(pitch_,b1,b2,Cp1,Cp2);
	return Cpp;
  }

  b1 = 4.7;
  b2 = 5.7;
  if( (pitch_ > b1)&&(pitch_ <= b2) ) {
	col = 3;
	Cp1 = CplArray[row][col];
	col = 4;
	Cp2 = CplArray[row][col];
    Cpp = Interpolate(pitch_,b1,b2,Cp1,Cp2);
	return Cpp;
  }

  b1 = 5.7;
  b2 = 6.7;
  if( (pitch_ > b1)&&(pitch_ <= b2) ) {
	col = 4;
	Cp1 = CplArray[row][col];
	col = 5;
	Cp2 = CplArray[row][col];
    Cpp = Interpolate(pitch_,b1,b2,Cp1,Cp2);
	return Cpp;
  }

  b1 = 6.7;
  b2 = 7.7;
  if( (pitch_ > b1)&&(pitch_ <= b2) ) {
	col = 5;
	Cp1 = CplArray[row][col];
	col = 6;
	Cp2 = CplArray[row][col];
    Cpp = Interpolate(pitch_,b1,b2,Cp1,Cp2);
	return Cpp;
  }

  b1 = 7.7;
  b2 = 8.7;
  if( (pitch_ > b1)&&(pitch_ <= b2) ) {
	col = 6;
	Cp1 = CplArray[row][col];
	col = 7;
	Cp2 = CplArray[row][col];
    Cp = Interpolate(pitch_,b1,b2,Cp1,Cp2);
	return Cp;
  }

  b1 = 8.7;
  b2 = 13.7;
  if( (pitch_ > b1)&&(pitch_ <= b2) ) {
	col = 7;
	Cp1 = CplArray[row][col];
	col = 8;
	Cp2 = CplArray[row][col];
    Cpp = Interpolate(pitch_,b1,b2,Cp1,Cp2);
	return Cpp;
  }

  b1 = 13.7;
  b2 = 18.7;
  if( (pitch_ > b1)&&(pitch_ <= b2) ) {
	col = 8;
	Cp1 = CplArray[row][col];
	col = 9;
	Cp2 = CplArray[row][col];
    Cpp = Interpolate(pitch_,b1,b2,Cp1,Cp2);
	return Cpp;
  }

  b1 = 18.7;
  b2 = 23.7;
  if( (pitch_ > b1)&&(pitch_ <= b2) ) {
	col = 9;
	Cp1 = CplArray[row][col];
	col = 10;
	Cp2 = CplArray[row][col];
    Cpp = Interpolate(pitch_,b1,b2,Cp1,Cp2);
	return Cpp;
  }

  b1 = 23.7;
  b2 = 28.7;
  if( (pitch_ > b1)&&(pitch_ <= b2) ) {
	col = 10;
	Cp1 = CplArray[row][col];
	col = 11;
	Cp2 = CplArray[row][col];
    Cpp = Interpolate(pitch_,b1,b2,Cp1,Cp2);
	return Cpp;
  }

  b1 = 28.7;
  b2 = 33.7;
  if( (pitch_ > b1)&&(pitch_ <= b2) ) {
	col = 11;
	Cp1 = CplArray[row][col];
	col = 12;
	Cp2 = CplArray[row][col];
    Cpp = Interpolate(pitch_,b1,b2,Cp1,Cp2);
	return Cpp;
  }
  return 1;
}


//calculate rayleigh wind speed
double ATurbine::CalcWindSpeed(double averageSpeed) {
	double test;
	double calcSpeed;
	double c;

	c = 2 * averageSpeed / FMath::Sqrt(pi);
	test = FMath::RandRange(0.0f, 1.0f);
	calcSpeed = c * FMath::Sqrt(-1 * FMath::Loge(test));

	return calcSpeed;
}


//power of function, cause the built in math one uses floats
//(WARNING DOESNT WORK FOR POWERS EQUALING:  0, NEGATIVE AND NOT WHOLE NUMBERS)
double ATurbine::PowerOf(double baseNum, int expNum) {
	switch(expNum){
		case 1:
			return baseNum;
		default:
			double answerVal = baseNum;
			for (int i = 1; i < expNum; i++) answerVal = answerVal * baseNum;
			return answerVal;
	}
}



//==========================================================================================================================================
//	VISUALIZATION
//==========================================================================================================================================

void ATurbine::Visualization(float deltaTime) {
	//rotate blades according to vals
	for (int i = 0; i < 3; i++) {
		float deltaRot = omega * deltaTime;
		blades[i]->AddActorWorldRotation(FRotator(deltaRot,0,0).Quaternion());
		//blades[i]->AddActorWorldRotation(FRotator(theta, 0, 0).Quaternion());
		// GEngine->AddOnScreenDebugMessage(-1, 40.0f, FColor::Yellow,
		// 	FString::Printf(TEXT("omega: %f time: %f"), omega, time));
	}

	//save snapshot
	//snapshots.Add(FSnapshot{time,windV,omega,Cp,power,tsr,beta});
}



//==========================================================================================================================================
//	AUTO-GENERATING OCTAVE FILES
//==========================================================================================================================================

void ATurbine::EndPlay(const EEndPlayReason::Type EndPlayReason) {
	Super::EndPlay(EndPlayReason);

	//get the file path
	FString logFile = FPaths::ProjectDir();
	FString fileName = "OctaveFiles/GeneratedPlots.m";
	logFile.Append(fileName);

	//add recorded vals

	int ticksRan = snapshots.Num();

	GEngine->AddOnScreenDebugMessage(-1, 40.0f, FColor::Red,
		FString::Printf(TEXT(">>>> EndPlay: number of logs = %d"), ticksRan));
	//time
	FString stringToSave = "times = [";
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"),snapshots[i].time));
	stringToSave.Append("];\n");
	//wind velocity
	stringToSave.Append("windVs = [");
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"),snapshots[i].windV));
	stringToSave.Append("];\n");
	//rotation speed
	stringToSave.Append("omegas = [");
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"),snapshots[i].omega));
	stringToSave.Append("];\n");
	//error
	stringToSave.Append("Cps = [");
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"),snapshots[i].Cp));
	stringToSave.Append("];\n");
	//power
	stringToSave.Append("powers = [");
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"),snapshots[i].power));
	stringToSave.Append("];\n");
	//tip speed ratio
	stringToSave.Append("tsrs = [");
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"),snapshots[i].tsr));
	stringToSave.Append("];\n");
	//beta
	stringToSave.Append("betas = [");
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"),snapshots[i].beta));
	stringToSave.Append("];\n");
	
	//plot graphs
	//windV / time
	// Mod 14-11-21 CBP autoscale of time axis
	stringToSave.Append("maxTimes = floor(max(times));\n");
	
	stringToSave.Append("subplot(3,2,1)\n");
	stringToSave.Append("plot(times(:),windVs(:));\n");
	stringToSave.Append("xlabel('Time');\n");
	stringToSave.Append("ylabel('windV');\n");
	//CODERS
	stringToSave.Append("line([0, max(times)], [11.7, 11.7]);\n");
	stringToSave.Append("axis([0,floor(max(times)),0,max(windVs*1.1)]);\n");
	//omega / time
	stringToSave.Append("subplot(3,2,2)\n");
	stringToSave.Append("plot(times(:),omegas(:));\n");
	stringToSave.Append("xlabel('Time');\n");
	stringToSave.Append("ylabel('omega');\n");
	// CODERS
	stringToSave.Append("line([0, max(times)], [3.86, 3.86]);\n");
	stringToSave.Append("axis([0,floor(max(times)),0,max(omegas*1.1)]);\n");

	//error / time
	stringToSave.Append("subplot(3,2,3)\n");
	stringToSave.Append("plot(times(:),Cps(:));\n");
	stringToSave.Append("xlabel('Time');\n");
	stringToSave.Append("ylabel('Cp');\n");
	//CODERS
	stringToSave.Append("line([0, max(times)], [0.463, 0.463]);\n");
	stringToSave.Append("axis([0,floor(max(times)),0,max(Cps*1.1)]);\n");
	//power / time
	stringToSave.Append("subplot(3,2,4)\n");
	stringToSave.Append("plot(times(:),powers(:));\n");
	stringToSave.Append("xlabel('Time');\n");
	stringToSave.Append("ylabel('power');\n");
	// CODERS
	stringToSave.Append("line([0, max(times)], [600000, 600000]);\n");
	stringToSave.Append("axis([0,floor(max(times)),0,max(powers*1.1)]);\n");
	//tip speed ratio / time
	stringToSave.Append("subplot(3,2,5)\n");
	stringToSave.Append("plot(times(:),tsrs(:));\n");
	stringToSave.Append("xlabel('Time');\n");
	stringToSave.Append("ylabel('ts ratio');\n");
	//CODERS
	//They should work out how to get tsrOpt value in the string, so we don't need to hard-code this.
	stringToSave.Append("line([0, max(times)], [6.6, 6.6]);\n");
	stringToSave.Append("axis([0,floor(max(times)),0,max(tsrs*1.1)]);\n");
	//beta / time
	stringToSave.Append("subplot(3,2,6)\n");
	stringToSave.Append("plot(times(:),betas(:));\n");
	stringToSave.Append("xlabel('Time');\n");
	stringToSave.Append("ylabel('beta');\n");
	stringToSave.Append("axis([0,floor(max(times)),0,max(betas*1.1)]);\n");

	//generate the octave file
	FFileHelper::SaveStringToFile(stringToSave, *logFile);

	GetWorldTimerManager().ClearTimer(TimerHandle);
}