// ============================================================================
// E:\aUE4\FBO_CBP_TurbineSim
// CBP 13-11-21
// Based on Dean Sisman's code
// ============================================================================

#pragma once

#define NEQN 1

#include "CoreMinimal.h"
#include "GameFramework/Actor.h"
#include "Turbine.generated.h"

UENUM() 
enum WindType {
	steady	    UMETA(DisplayName = "Steady"),
	stepped		UMETA(DisplayName = "Stepped"),
	series	    UMETA(DisplayName = "Time Series"),
};

UCLASS()
class TURBINESIM2_API ATurbine : public AActor
{
	GENERATED_BODY()
	

public:	
	// Sets default values for this actor's properties
	ATurbine();

protected:
	// Called when the game starts or when spawned
	virtual void BeginPlay() override;
	virtual void EndPlay(const EEndPlayReason::Type EndPlayReason) override;

public:	
	// Called every frame
	//virtual void Tick(float DeltaTime) override;
	FTimerHandle TimerHandle;
	void ComputeDT();
	int itn;
	void LogData();
	bool DEBUG = false;

//private methods
private:
	void SpawnBlades();
	void ReadCpLambdaFile();
	void Computation();
	void Visualization(float deltaTime);
	double LookUpInCpLambdaArray(double tsr_, double pitch_);
	double Interpolate(double beta_, double beta1_, double beta2_, double Cp1_, double Cp2_);
	double CalcWindSpeed(double averageSpeed);
	void Region2();
	double PowerOf(double baseNum, int expNum);
	double SteppedWindSpeed();

//internal variables
private:
/*==========================#=======================================#===========================*/
/*	code					#	meaning								#	default/starting value	*/
/*==========================#=======================================#===========================*/
	bool bRunning = true;
	int region;      		/*	which region is currently active	#	-						*/
	int prevRegion;
	double windV = 0;  		/*	wind velocity m/s					#	0						*/
	double windVRayleigh;  	/*	calculated wind v					#	-						*/
	double omega;    		/*	angular speed of turbine shaft		#	-						*/
	double omegaRated;  	/*	rated speed of turbine shaft		#	-						*/
	double theta = 0;    	/*	angle of turbine shaft				#	0						*/
	//double coeffPow; 		/*	power coefficient					#	-						*/
	double tsr;      		/*	tip-speed ratio						#	-						*/
	double power;    		/*	power output						#	-						*/
	double torqueW;  		/*	torque on shaft due to wind			#	-						*/
	double torqueG;  		/*	torque of shaft due to generator	#	-						*/
	double torqueB = 0.0; 	/*	torque of shaft due to e-brake		#	0.0						*/
	double error;    		/*	difference between two torques		#	-						*/
	double beta;     		/*	pitch of the rotor blades			#	-						*/
	double deltaBeta = 0;	/*	change in pitch of the rotor blades	#	0						*/
	double cpMax = 0.4633;  /*	maximum power coefficient			#	0.4633					*/
	double tsrOpt = 6.6;  	/*	optimal value of tip-speed ratio	#	6.6						*/
	double time = 0;	  	/*	tracks time while sim is running	#	0						*/
	double pi = 3.141592653;/*	tracks time while sim is running	#	3.141592653				*/
	int tickTracker = 0;	/*	the amount of ticks the sim has ran	#	0						*/
	double stepTime = 0.0;	/*	the time since the last wind step	#	0.0						*/
	double Cp;				/* power coefficient												*/	
	double scaling = 50;	/*	rotation scaling					#	50.0					*/
/*==========================#=======================================#===========================*/
	bool region3MessageShown;

//vars exposed to editor
public:
/*======================================#=======================================#===========================*/
/*	code								#	meaning								#	default/starting value	*/
/*======================================#=======================================#===========================*/
	UPROPERTY(EditAnywhere,Category = "Wind")
		TEnumAsByte<WindType> windType;
	UPROPERTY(EditAnywhere, Category = "Simulation Engine")
		double stopTime = 500;

	UPROPERTY(EditAnywhere)				/*										#							*/
	  double R = 20;					/*	radius of blades					#	20						*/
	UPROPERTY(EditAnywhere)				/*										#							*/
	  double rho = 1.2;					/*	air density							#	1.2						*/
	UPROPERTY(EditAnywhere)				/*										#							*/
	  double J = 613301;				/*	moment of intertia					#	613301  				*/
//	UPROPERTY(EditAnywhere)				/*										#							*/
//	  bool useSteppedWindSpeed = false;	/*	whether to use stepped wind speed	#	false					*/
	UPROPERTY(EditAnywhere, Category = "Wind|Stepped")
		double windStepTimeInterval = 20.0;
	UPROPERTY(EditAnywhere, Category = "Wind|Stepped")
		double windStepHeight = 1.0;
	UPROPERTY(EditAnywhere, Category = "Wind|Time Series")
		double averageWindSpeed = 5.0;
	UPROPERTY(EditAnywhere)				/*										#							*/
	  double vCutIn = 4.0;				/*	cut in wind speed					#	4.0						*/
	UPROPERTY(EditAnywhere)				/*										#							*/
	  double vRated = 11.7;				/*	rated wind speed					#	11.7					*/
	UPROPERTY(EditAnywhere)				/*										#							*/
	  double vCutOut = 25;				/*	cut out wind speed					#	25						*/
	UPROPERTY(EditAnywhere)				/*										#							*/
	  double omegaInit = 2;				/*	initial rotation speed				#	2						*/
	
	UPROPERTY(EditAnywhere, Category = "Wind|Stepped and Steady")	
	  double windVInit = 8;	


	UPROPERTY(EditAnywhere)				/*										#							*/
	  double region2Beta = 3.7;			/*	rotor blade pitch in region 2		#	3.7						*/
	UPROPERTY(EditAnywhere, Category = "Simulation Engine")				/*										#							*/
	  double dT = 0.01;					/*	the timestep						#	0.01					*/
	UPROPERTY(EditAnywhere)				/*										#							*/
	double Kp = 2.0;					/*										#	2.0						*/
	//UPROPERTY(EditAnywhere)				/*										#							*/
	//double Kd = 0.0;					/*										#	0.0						*/
	UPROPERTY(EditAnywhere, Category = "Simulation Engine")
		int speedUp = 10;
	//UPROPERTY(EditAnywhere)				/*										#							*/
	// bool realTimeVisualization = false;/*	whether to render in real time		#	false					*/
/*======================================#=======================================#===========================*/
	UPROPERTY(EditAnywhere, Category = "Simulation Engine")
		double logInterval = 0.1;
//arrays
private:
	TArray<class ABlade*> blades;
	TArray<FCplArray> CplArray;
	TArray<FSnapshot> snapshots;

//solver parameters for runge-kutta-fehlberg
private:
	double y[NEQN];
	double dydt[NEQN];
	double t;
	double tout;
	double relerr;
	double abserr;
	int flag;
};



//the struct for the 2d array needed for the cpl
USTRUCT()
struct FCplArray{
    GENERATED_BODY()
	public:
		TArray<double> cplVal;
		double operator[] (int32 i) {
			return cplVal[i];
		}
		void Add(double value) {
			cplVal.Add(value);
		}
};



//struct for recording the simulation
USTRUCT()
struct FSnapshot {
	GENERATED_BODY()
	double time, windV, omega, Cp, power, tsr, beta;
};